# -*- coding: utf-8 -*-

require 'isono'

module Dcmgr
  module NodeModules
    class EventHook < Isono::NodeModules::Base
      include Dcmgr::Logger

      initialize_hook do
        @worker_thread = Isono::ThreadPool.new(1, 'EventHook')

        event = Isono::NodeModules::EventChannel.new(node)
        event.subscribe('hva/instance_terminated', '#') do |args|
          logger.info("event caught: #{args.inspect}")
          instance_id = args[0]
          @worker_thread.pass {
            Sequel::DATABASES.first.transaction do
              instance = Models::Instance[instance_id]
              network_vif_uuids = instance.network_vif.collect{ |nv| nv.canonical_uuid }
              load_balancers = Models::LoadBalancerTarget.get_load_balancers(network_vif_uuids)
              load_balancers.each { |lb|
                myinstance.unregister_instance(lb.canonical_uuid, network_vif_uuids)
                myinstance.update_load_balancer(lb.canonical_uuid, network_vif_uuids)
              }
              # Delete instance log monitoring.
              myinstance.delete_log_monitorings(instance)
            end
          }
        end
      end

      terminate_hook do
        @worker_thread.shutdown
      end

      def unregister_instance(load_balancer_id, network_vif_uuids)
        begin
          lb = Models::LoadBalancer[load_balancer_id]
          raise "Unknown Load Balancer ID: #{load_balancer_id}"if lb.nil?

          network_vif_uuids.each {|vif_uuid|
            lb.remove_instance_security_group(vif_uuid)
          }

          lbt = lb.remove_targets(network_vif_uuids)
          logger.info("Unregister interface #{network_vif_uuids.to_s} from Load Balancer #{lb.id}.")
        rescue ::Exception => e
          logger.error("Error occured. [load_balancer_id: #{load_balancer_id}]")
          logger.error(e)
        end
      end

      def update_load_balancer(load_balancer_id, exclude_vifs)
        lb = Models::LoadBalancer[load_balancer_id]
        raise "Unknown Load Balancer ID: #{load_balancer_id}" if lb.nil?

        load_balancer_config = lb.get_reload_config({
          :servers => lb.get_target_servers(:exclude_vifs => exclude_vifs)
        })

        Sequel::DATABASES.first.after_commit do
          Messaging::LoadBalancer.update_load_balancer_config(load_balancer_config)
          logger.info("Update load balancer: #{load_balancer_id}")
        end

      end

      def delete_log_monitorings(instance)
        logger.info("Delete log monitorings: #{instance.canonical_uuid}")
        monitorings = instance.delete_log_monitorings
        unless monitorings.empty?
          Sequel::DATABASES.first.after_commit do
            Dcmgr.messaging.submit("log-alarm-registry.#{instance.host_node.node_id}", 'delete_alarms')
            logger.info("Messaging submit: #{instance.host_node.node_id}, delete_alarms")
          end
        end

      end
    end
  end
end
